from dataclasses import dataclass
from typing import Any, Optional

from . import Settings


@dataclass
class TrainerConf:
    _target_: str = "pytorch_lightning.trainer.Trainer"
    logger: Any = (
        True  # Union[LightningLoggerBase, Iterable[LightningLoggerBase], bool]
    )
    enable_checkpointing: bool = True
    callbacks: Any = None  # Optional[List[Callback]]
    default_root_dir: Optional[str] = None
    gradient_clip_val: float = 0
    process_position: int = 0
    num_nodes: int = 1
    # num_processes: int = 1
    # gpus: Any = None  # Union[int, str, List[int], NoneType]
    devices: Any = None  # Union[int, str, List[int], NoneType]
    auto_select_gpus: bool = False
    # tpu_cores: Any = None  # Union[int, str, List[int], NoneType]
    log_gpu_memory: Optional[str] = None
    overfit_batches: Any = 0.0  # Union[int, float]
    track_grad_norm: Any = -1  # Union[int, float, str]
    check_val_every_n_epoch: int = 1
    fast_dev_run: Any = False  # Union[int, bool]
    accumulate_grad_batches: Any = 1  # Union[int, Dict[int, int], List[list]]
    max_epochs: int = 1000
    min_epochs: int = 1
    max_steps: int = -1
    min_steps: Optional[int] = None
    limit_train_batches: Any = 1.0  # Union[int, float]
    limit_val_batches: Any = 1.0  # Union[int, float]
    limit_test_batches: Any = 1.0  # Union[int, float]
    val_check_interval: Any = 1.0  # Union[int, float]
    log_every_n_steps: int = 50
    accelerator: Any = None  # Union[str, Accelerator, NoneType]
    sync_batchnorm: bool = False
    precision: int = 32
    # weights_summary: Optional[str] = "top"
    # weights_save_path: Optional[str] = None
    num_sanity_val_steps: int = 2
    # resume_from_checkpoint: Any = None  # Union[str, Path, NoneType]
    profiler: Any = None  # Union[BaseProfiler, bool, str, NoneType]
    benchmark: bool = False
    deterministic: bool = False
    # reload_dataloaders_every_epoch: bool = False
    auto_lr_find: Any = False  # Union[bool, str]
    replace_sampler_ddp: bool = True
    auto_scale_batch_size: Any = False  # Union[str, bool]
    plugins: Any = None  # Union[str, list, NoneType]
    amp_backend: str = "native"
    # amp_level: str = "O2"
    move_metrics_to_cpu: bool = False
    detect_anomaly: bool = False


@dataclass
class LightningConf(TrainerConf):
    deterministic: bool = True
    accelerator: str = "gpu"
    devices: int = 1


@dataclass
class LightningSettings(Settings):
    # Lightning trainer settings
    pl: LightningConf = LightningConf()

    # Additional experiment settings
    validate_before_training: bool = True
    save_checkpoints: bool = True



import os
from datetime import datetime
from pathlib import Path

import coolname
from dotenv import load_dotenv

load_dotenv()


def setup_rundir():
    """
    Create a separate working directory under `$RESULTS_DIR/$WANDB_PROJECT` with a randomly generated run name.
    """
    date = datetime.now().strftime("%Y%m%d-%H%M")
    name = coolname.generate_slug(2)  # type: ignore
    os.environ['RUN_NAME'] = f'{date}-{name}'

    results_root = f'{os.getenv("RESULTS_DIR")}/{os.getenv("WANDB_PROJECT")}'
    if os.getenv('RUN_MODE', '').lower() == 'debug':
        run_dir = f'{results_root}/debug/{os.getenv("RUN_NAME")}'
        completed_run_dir = run_dir
        os.environ['WANDB_MODE'] = 'disabled'
    else:
        run_dir = f'{results_root}/running/{os.getenv("RUN_NAME")}'
        completed_run_dir = f'{results_root}/completed/{os.getenv("RUN_NAME")}'

    os.makedirs(run_dir, exist_ok=True)
    os.environ['RUN_DIR'] = run_dir
    os.environ['COMPLETED_RUN_DIR'] = completed_run_dir


def finish_rundir():
    """
    Move the run directory to completed subdirectory.
    """
    run_dir = Path(os.getenv('RUN_DIR', ''))
    completed_run_dir = Path(os.getenv('COMPLETED_RUN_DIR', ''))

    if run_dir != completed_run_dir:
        assert run_dir.exists() and run_dir.is_dir()
        assert not completed_run_dir.exists()
        os.makedirs(completed_run_dir.parent, exist_ok=True)
        run_dir.rename(completed_run_dir)



import re
from typing import Any, Optional, Type

from rich import box
from rich.console import Console
from rich.progress import (BarColumn, Progress, SpinnerColumn, TaskID,
                           TimeElapsedColumn, TimeRemainingColumn)
from rich.table import Table
from torch import Tensor

from avalanche.evaluation.metric_results import MetricValue, TensorImage
from avalanche.evaluation.metric_utils import phase_and_task, stream_type
from avalanche.logging import StrategyLogger
from avalanche.training.strategies.base_strategy import BaseStrategy

UNSUPPORTED_TYPES: tuple[Type[Any]] = (TensorImage, )


class RichLogger(StrategyLogger):
    def __init__(self, ignore_metrics: Optional[list[str]] = None, start_from_epoch_one: bool = True,
                 start_from_experience_one: bool = True):
        super().__init__()
        self.metric_vals: dict[str, Any] = {}
        self.console = Console()

        self.ignore_metrics = ignore_metrics or []
        self.start_from_epoch_one = start_from_epoch_one
        self.start_from_experience_one = start_from_experience_one

        self._pbar: Optional[Progress] = None
        self._pbar_task: Optional[TaskID] = None

        self._reset_table()

    def _reset_table(self):
        self._table = Table(show_header=False, show_edge=True, box=box.ROUNDED)
        self._table.add_column('Metric', style='white', min_width=40)
        self._table.add_column('Value', style='bold magenta')

    def log_metric(self, metric_value: 'MetricValue', callback: str) -> None:
        name = metric_value.name
        x = metric_value.x_plot
        val = metric_value.value
        self.metric_vals[name] = (name, x, val)

    def _val_to_str(self, value: Any):
        if isinstance(value, Tensor):
            return '\n' + str(value)
        elif isinstance(value, float):
            return f'{value:.4f}'
        else:
            return str(value)

    def write_metrics(self):
        sorted_vals = sorted(self.metric_vals.values(),
                             key=lambda x: x[0])
        for name, _, val in sorted_vals:
            ignored = False
            for regex in self.ignore_metrics:
                if re.match(regex, name):
                    ignored = True
                    break
            if isinstance(val, UNSUPPORTED_TYPES) or ignored:
                continue
            val = self._val_to_str(val)
            self._table.add_row(name, val)

    def flush_metrics(self):
        if self._table.row_count:
            self.console.print(self._table)
        self._reset_table()

    def before_training(self, strategy: BaseStrategy,
                        metric_values: list[MetricValue], **kwargs: Any):
        super().before_training(strategy, metric_values, **kwargs)  # type: ignore
        self.console.print('\n[bold yellow]» :runner: Start of training phase...')

    def before_training_exp(self, strategy: BaseStrategy,
                            metric_values: list[MetricValue], **kwargs: Any):
        super().before_training_exp(strategy, metric_values, **kwargs)  # type: ignore
        self._on_start(strategy)

    def before_training_epoch(self, strategy: BaseStrategy,
                              metric_values: list[MetricValue], **kwargs: Any):
        super().before_training_epoch(strategy, metric_values, **kwargs)  # type: ignore
        self._progress.update(self._main_task, total=len(strategy.dataloader))  # type: ignore

        epoch, experience = self.get_epoch_experience(strategy)
        total = strategy.train_epochs

        self.console.rule(f'[bold yellow]Epoch {epoch} of {total} [cyan]|[yellow] Experience {experience}')
        self.console.print(f'[bold cyan]» :runner: Training (epoch {epoch})...')

        self._on_start(strategy)

    def after_training_iteration(self, strategy: BaseStrategy,
                                 metric_values: list[MetricValue], **kwargs: Any):
        self._progress.advance(self._main_task)
        self._progress.refresh()
        super().after_training_iteration(strategy, metric_values, **kwargs)  # type: ignore

    def after_training_epoch(self, strategy: BaseStrategy,
                             metric_values: list[MetricValue], **kwargs: Any):
        self._end_progress()
        super().after_training_epoch(strategy, metric_values, **kwargs)  # type: ignore
        self.write_metrics()
        self.flush_metrics()
        self.metric_vals = {}
        # self.console.print(
        # f'[bold green]» :checkered_flag: Training epoch {strategy.epoch} completed. :checkered_flag:')

    def after_training(self, strategy: BaseStrategy,
                       metric_values: list[MetricValue], **kwargs: Any):
        super().after_training(strategy, metric_values, **kwargs)  # type: ignore
        self.console.print('[bold green]» :checkered_flag: End of training phase. :checkered_flag:')

    def before_eval(self, strategy: BaseStrategy,
                    metric_values: list[MetricValue], **kwargs: Any):
        super().before_eval(strategy, metric_values, **kwargs)  # type: ignore

        epoch, _ = self.get_epoch_experience(strategy)
        self.console.print(f'\n[bold yellow]» :magnifying_glass_tilted_left: Evaluation (epoch {epoch})...')

    def before_eval_exp(self, strategy: BaseStrategy,
                        metric_values: list[MetricValue], **kwargs: Any):
        super().before_eval_exp(strategy, metric_values, **kwargs)  # type: ignore
        self._on_start(strategy)
        self._progress.update(self._main_task, total=len(strategy.dataloader))  # type: ignore

    def after_eval_iteration(self, strategy: BaseStrategy,
                             metric_values: list[MetricValue], **kwargs: Any):
        self._progress.advance(self._main_task)
        self._progress.refresh()
        super().after_eval_iteration(strategy, metric_values, **kwargs)  # type: ignore

    def after_eval_exp(self, strategy: BaseStrategy,
                       metric_values: list[MetricValue], **kwargs: Any):
        self._end_progress()
        super().after_eval_exp(strategy, metric_values, **kwargs)  # type: ignore

        self.write_metrics()
        self.metric_vals = {}

    def after_eval(self, strategy: BaseStrategy,
                   metric_values: list[MetricValue], **kwargs: Any):
        super().after_eval(strategy, metric_values, **kwargs)  # type: ignore
        self.write_metrics()
        self.flush_metrics()
        self.metric_vals = {}
        # self.console.print('[bold yellow]» :checkered_flag: [green]End of eval phase. :checkered_flag:')

    def get_epoch_experience(self, strategy: BaseStrategy):
        if strategy.epoch is None:
            return None, None

        epoch = strategy.epoch + 1 if self.start_from_epoch_one else strategy.epoch
        experience: int = strategy.experience.current_experience  # type: ignore
        if self.start_from_experience_one:
            experience += 1

        return epoch, experience

    def _on_start(self, strategy: BaseStrategy):
        action_name = 'train' if strategy.is_training else 'eval'
        _, experience = self.get_epoch_experience(strategy)

        task_id = phase_and_task(strategy)[1]
        stream = stream_type(strategy.experience)  # type: ignore
        if task_id is None:
            description = f'[cyan] | [bold white]{action_name}[normal] » {stream} » exp {experience}'
        else:
            description = f'[cyan] | [bold white]{action_name}[normal] » {stream} » exp {experience} (task {task_id})'
        self._progress.update(self._main_task, description=description)

    @property
    def _progress(self):
        if self._pbar is None:
            columns = [
                "  [progress.percentage][{task.completed} / {task.total}]",
                BarColumn(),
                "[progress.percentage]{task.percentage:>3.0f}%",
                SpinnerColumn(),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
                "[progress.description]{task.description}",
            ]

            self._pbar = Progress(*columns, console=self.console, transient=True)
            self._pbar_task = self._pbar.add_task('', total=0)
            self._pbar.start()

        return self._pbar

    @property
    def _main_task(self):
        assert self._pbar_task is not None
        return self._pbar_task

    def _end_progress(self):
        if self._pbar is not None:
            self._pbar.stop()
            self._pbar = None
            self._pbar_task = None



from contextlib import contextmanager
from typing import Any, Optional, cast

import pytorch_lightning as pl
from pytorch_lightning.loggers.base import LightningLoggerBase, LoggerCollection
from pytorch_lightning.loggers.wandb import WandbLogger

from wandb.sdk.wandb_run import Run


class LightningModuleWithWandb(pl.LightningModule):
    """
    Vanilla LightningModule with WandbLogger experiment pinned to self.wandb.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.wandb: Optional[Run] = None

    @property
    def logger(self) -> LightningLoggerBase:
        """Reference to the logger object in the Trainer."""
        return self.trainer.logger if self.trainer else None  # type: ignore

    @property
    def epoch(self) -> int:
        return self.current_epoch + 1 if not self.trainer.sanity_checking else self.current_epoch  # type: ignore

    def log_wandb(self, data: dict[str, Any], step: int):
        if self.wandb is not None:
            self.wandb.log(data, step=step)

    def on_fit_start(self) -> None:
        if isinstance(self.logger, LoggerCollection):
            for logger in self.logger:
                if isinstance(logger, WandbLogger):
                    self.wandb = cast(Run, logger.experiment)  # type: ignore
        elif isinstance(self.logger, WandbLogger):
            self.wandb = cast(Run, self.logger.experiment)  # type: ignore

    @contextmanager
    def no_train(self):
        training = self.training
        self.train(False)
        yield
        self.train(training)



import os
from abc import abstractmethod
from pathlib import Path
from typing import Any, Optional, cast

import hydra
import pytorch_lightning as pl
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import WandbLogger
from wandb.sdk.wandb_run import Run

from ..configs import get_tags
from ..configs.cfg.lightning import LightningSettings
from ..core import Experiment
from ..utils.logging import error, info, info_bold


class LightningExperiment(Experiment):
    def before_entry(self) -> None:
        self.cfg: LightningSettings

        self.wandb_logger: WandbLogger
        self.datamodule: Optional[pl.LightningDataModule] = None
        self.system: Optional[pl.LightningModule] = None
        self.callbacks: list[Any] = []
        self.trainer: Optional[pl.Trainer] = None

        self.wandb_logger = WandbLogger(
            project=os.getenv("WANDB_PROJECT"),
            entity=os.getenv("WANDB_ENTITY"),
            name=os.getenv("RUN_NAME"),
            save_dir=str(Path(os.getenv("RUN_DIR", "."))),
        )

        # Init logger from source dir (code base) before switching to run dir (results)
        self.wandb_logger.experiment  # type: ignore
        self.run: Run = self.wandb_logger.experiment  # type: ignore

    def main(self) -> None:
        tags = get_tags(cast(DictConfig, self.root_cfg))
        self.run.tags = tags
        self.run.notes = str(self.root_cfg.notes)
        self.wandb_logger.log_hyperparams(OmegaConf.to_container(self.root_cfg.cfg, resolve=True))  # type: ignore

        Path(self.root_cfg.data_dir).mkdir(parents=True, exist_ok=True)
        self.setup_datamodule()
        assert (
            self.datamodule is not None
        ), "A LightningDataModule should be assigned to self.datamodule inside setup_datamodule()."

        self.setup_system()
        assert (
            self.system is not None
        ), "A main LightningModule should be assigned to self.system inside setup_system()."
        info_bold(f"System architecture:")
        info(str(self.system))

        self.setup_callbacks()

        self.setup_trainer()
        assert self.trainer is not None

        self.fit()

    @abstractmethod
    def setup_datamodule(self) -> None:
        pass
        # TODO: Add datamodule instantiation?
        # datamodule: LightningDataModule = hydra.utils.instantiate(
        #     cfg.experiment.datamodule,
        #     batch_size=cfg.experiment.batch_size,
        #     seed=cfg.experiment.seed,
        #     shuffle=cfg.experiment.shuffle,
        #     num_workers=cfg.experiment.num_workers
        # )

    @abstractmethod
    def setup_system(self) -> None:
        pass

    def setup_callbacks(self) -> None:
        if self.cfg.save_checkpoints:
            checkpointer = ModelCheckpoint(
                dirpath="checkpoints",
                filename="epoch{epoch:02d}",
                auto_insert_metric_name=False,
                every_n_epochs=1,
                save_on_train_epoch_end=True,
                save_weights_only=True,
            )

            self.callbacks.append(checkpointer)

    def setup_trainer(self) -> None:
        num_sanity_val_steps = -1 if self.cfg.validate_before_training else 0

        # TODO: Add resume support
        # resume_path = get_resume_checkpoint(cfg, wandb_logger)
        # if resume_path is not None:
        #     log.info(f'[bold yellow]\\[checkpoint] [bold white]{resume_path}')

        info_bold(f"Overriding cfg.pl settings with derived values:")
        # info(f' >>> resume_from_checkpoint = {resume_path}')
        info(f" >>> num_sanity_val_steps = {num_sanity_val_steps}")
        info("")

        self.trainer = hydra.utils.instantiate(  # type: ignore
            self.cfg.pl,
            logger=self.wandb_logger,
            callbacks=self.callbacks,
            enable_checkpointing=self.cfg.save_checkpoints,
            num_sanity_val_steps=num_sanity_val_steps,
            #     resume_from_checkpoint=resume_path,
        )

    def fit(self) -> None:
        assert self.system
        self.trainer.fit(self.system, datamodule=self.datamodule)  # type: ignore

    def finish(self) -> None:
        if self.trainer.interrupted:  # type: ignore
            error(f"Training interrupted.")
            self.run.finish(exit_code=255)
        else:
            self.run.finish()



from typing import Any, Optional, cast

import numpy as np
from matplotlib.figure import Figure
from numpy import array
from PIL.Image import Image
from torch import Tensor

import wandb
import wandb.viz
from avalanche.benchmarks.scenarios.new_classes.nc_scenario import NCExperience
from avalanche.evaluation.metric_results import (AlternativeValues,
                                                 MetricValue, TensorImage)
from avalanche.logging.strategy_logger import StrategyLogger
from avalanche.training.strategies.base_strategy import BaseStrategy


class WandbEpochLogger(StrategyLogger):
    def __init__(self,
                 project: Optional[str] = None,
                 entity: Optional[str] = None,
                 name: Optional[str] = None,
                 dir: Optional[str] = None,
                 config: Optional[dict[str, Any]] = None,
                 tags: Optional[list[str]] = None,
                 notes: Optional[str] = None,
                 group: Optional[str] = None,
                 start_from_epoch_one: bool = True,
                 start_from_experience_one: bool = True
                 ):
        super().__init__()

        self.start_from_epoch_one = start_from_epoch_one
        self.start_from_experience_one = start_from_experience_one

        self.global_epoch = 1 if self.start_from_epoch_one else 0
        self.experience_epoch = 1 if self.start_from_epoch_one else 0

        self.current_experience = 1 if self.start_from_experience_one else 0

        self.wandb = wandb
        self.wandb.init(project=project, entity=entity, name=name, dir=dir,  # type: ignore
                        config=config, tags=tags, notes=notes, group=group)

    def log_metric(self, metric_value: MetricValue, callback: str):
        name = metric_value.name
        value = metric_value.value

        supported_types = (Image, Tensor, TensorImage, Figure,
                           float, int, self.wandb.Histogram, self.wandb.viz.CustomChart)

        if isinstance(value, AlternativeValues):
            value = value.best_supported_value(*supported_types)

        if not isinstance(value, supported_types):
            return  # Unsupported type

        if isinstance(value, Image):
            value = self.wandb.Image(value)
        elif isinstance(value, Tensor):
            value = np.histogram(value.view(-1).numpy())
            value = self.wandb.Histogram(np_histogram=value)
        elif isinstance(value, TensorImage):
            value = self.wandb.Image(np.moveaxis(array(value), 0, -1))

        self.wandb.log({  # type: ignore
            name: value,
            'epoch': self.global_epoch - 1 if 'eval' in callback else self.global_epoch,
            'experience_epoch': self.experience_epoch - 1 if 'eval' in callback else self.experience_epoch,
            'current_experience': self.current_experience,
        }, step=self.global_epoch - 1 if 'eval' in callback else self.global_epoch)

    def before_training(self, strategy: BaseStrategy, metric_values: list[MetricValue], **kwargs: Any):
        super().before_training(strategy, metric_values, **kwargs)  # type: ignore

    def after_training_epoch(self, strategy: BaseStrategy, metric_values: list[MetricValue], **kwargs: Any):
        super().after_training_epoch(strategy, metric_values, **kwargs)  # type: ignore

        self.global_epoch += 1
        self.experience_epoch += 1

    def before_training_exp(self, strategy: BaseStrategy, metric_values: list[MetricValue], **kwargs: Any):
        super().before_training_exp(strategy, metric_values, **kwargs)  # type: ignore

        experience = cast(NCExperience, strategy.experience)  # type: ignore

        self.current_experience = experience.current_experience
        if self.start_from_experience_one:
            self.current_experience += 1
        self.experience_epoch = 1 if self.start_from_epoch_one else 0



# Using redundant module aliases for public export
# https://github.com/microsoft/pyright/blob/master/docs/typed-libraries.md#library-interface

from .cfg import Settings as Settings
from .register import RootConfig as RootConfig
from .register import get_tags as get_tags
from .register import register_configs as register_configs



from dataclasses import dataclass
from typing import List, Optional


# Experiment settings validation schema & default values
@dataclass
class Settings:
    # ----------------------------------------------------------------------------------------------
    # General experiment settings
    # ----------------------------------------------------------------------------------------------
    # wandb tags
    _tags_: Optional[List[str]] = None

    # Seed for all random number generators
    seed: int = 1
